/******************************************************************************
 * Copyright (c) 2010, 2015  Ericsson AB
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 * Michael Josenhans
 ******************************************************************************/
//
//  File:               SocketCAN_PT.cc
//  Description:        SocketCAN_PT test port source
#include "SocketCAN_PT.hh"
#include "SocketCAN_PortType.hh"

#include <algorithm>
#include <iostream>
#include <string>
#include "linux/can.h"
#include "linux/can/bcm.h"
#include "linux/can/error.h"
#include "linux/can/raw.h"
#include "linux/if.h"
#include "memory.h"
#include "stdio.h"
#include "sys/socket.h"
#include "sys/ioctl.h"

#define DEFAULT_NUM_SOCK	      10
#define BCM_FRAME_BUFFER_SIZE    256

namespace SocketCAN__PortType {

SocketCAN__PT_PROVIDER::SocketCAN__PT_PROVIDER(const char *par_port_name) :
		PORT(par_port_name), debugging(true), target_fd(-1) {
	sock_list = NULL;
	num_of_sock = 0;
	sock_list_length = 0;

	operation_mode = false;
}

SocketCAN__PT_PROVIDER::~SocketCAN__PT_PROVIDER() {
	Free(sock_list);
}

void SocketCAN__PT_PROVIDER::set_parameter(const char * parameter_name,
		const char * parameter_value) {
	log("entering SocketCAN__PT_PROVIDER::set_parameter(%s, %s)",
			parameter_name, parameter_value);

	log("leaving SocketCAN__PT_PROVIDER::set_parameter(%s, %s)", parameter_name,
			parameter_value);
}

/*void SocketCAN__PT_PROVIDER::Handle_Fd_Event(int fd, boolean is_readable,
 boolean is_writable, boolean is_error) {}*/

void SocketCAN__PT_PROVIDER::Handle_Fd_Event_Error(int /*fd*/) {

}

void SocketCAN__PT_PROVIDER::Handle_Fd_Event_Writable(int /*fd*/) {

}

void SocketCAN__PT_PROVIDER::Handle_Fd_Event_Readable(int sock) {
	log("entering SocketCAN__PT_PROVIDER::Handle_Fd_Event_Readable()");
	int res;

	for (int a = 0; a < sock_list_length; a++) {
		if ((sock == sock_list[a].fd) and (sock_list[a].status)) {
			switch (sock_list[a].protocol_family) {
			case CAN_RAW: {
				SocketCAN__Types::SocketCAN__receive__CAN__or__CAN__FD__frame parameters;

				char msg[65536];
				struct sockaddr_can addr;
				socklen_t len = sizeof(addr);

				// here the bigger frame the canfd_frame is used for reception of both.

				//ssize_t recvfrom(int sockfd, void *buf, size_t len, int flags,
				// struct sockaddr *src_addr, socklen_t *addrlen);
				ssize_t nbytes = recvfrom(sock, &msg, CANFD_MTU, 0,
						(struct sockaddr*) &addr, &len);

				switch (nbytes) {

				case CAN_MTU: {
					struct can_frame frame;
					&frame = &msg;
					struct ifreq ifr;
					ifr.ifr_ifindex = addr.can_ifindex;
					/* get interface name of the received CAN frame */
					if (res = ioctl(sock, SIOCGIFNAME, &ifr) < 0) {
						TTCN_warning(
								"Ioctl failed while receiving a CAN frame on socket: %d with interface index %d\n",
								sock, ifr.ifr_ifindex);
					}
					if (debugging) { //if debug
						TTCN_Logger::begin_event(TTCN_DEBUG);
						TTCN_Logger::log_event_str(prompt);
						TTCN_Logger::log_event("Received a CAN frame from interface %s of bytes %d and with payload length %d",
								ifr.ifr_name, nbytes, frame.can_dlc);
						TTCN_Logger::log_event("\n");
						TTCN_Logger::end_event();
					}
					// fd_frame.flags contains invalid data
					parameters.ifr().if__index() = addr.can_ifindex;
					parameters.id() = a;
					parameters.frame().can__frame().can__id() = frame.can_id;
					parameters.frame().can__frame().can__pdu() = OCTETSTRING(
							frame.can_dlc, frame);
					incoming_message(parameters);
				}
					break;

				case CANFD_MTU: {
					struct canfd_frame fd_frame;
					&fd_frame = &msg;

					struct ifreq ifr;
					ifr.ifr_ifindex = addr.can_ifindex;
					/* get interface name of the received CAN frame */
					if (res = ioctl(sock, SIOCGIFNAME, &ifr) < 0) {
						TTCN_warning(
								"Ioctl failed while receiving a CAN FD message on socket: %d with interface index %d\n",
								sock, ifr.ifr_ifindex);
					}
					if (debugging) { //if debug
						TTCN_Logger::begin_event(TTCN_DEBUG);
						TTCN_Logger::log_event_str(prompt);
						TTCN_Logger::log_event("Received a CAN FD frame from interface %s of bytes %d and with payload length %d",
								ifr.ifr_name, nbytes, fd_frame.len);
						TTCN_Logger::log_event("\n");
						TTCN_Logger::end_event();
					}
					// fd_frame.flags contains valid data
					parameters.ifr().if__index() = addr.can_ifindex;
					parameters.id() = a;
					parameters.frame().can__fd__frame().can__id() =
							fd_frame.can_id;
					parameters.frame().can__fd__frame().can__flags() =
							fd_frame.flags;
					parameters.frame().can__fd__frame().can__pdu() =
							OCTETSTRING(fd_frame.len, fd_frame);
					incoming_message(parameters);
				}
					break;

				default: {
					if (nbytes <= 0) {
						//there is an empty message, or error in receive
						//remove the socket
						TTCN_warning(
								"Closing socket % with interface index %d due to an empty message or error in reception\n",
								sock, addr.can_ifindex);
						std::cout << "close_fd" << sock << std::endl;
						sock_list[a].status = 0;
						sock_list[a].protocol_family = 0;
						num_of_sock--;
						Handler_Remove_Fd_Read (close_fd);
						close(sock);
					} else {
						fprintf(stderr, "read: invalid CAN / CAN FD frame\n");
						TTCN_warning(
								"Closing socket % with interface index %d due reception of an invalid CAN / CAN FD frame\n",
								sock, addr.can_ifindex);
						//there has been an invalid CAN / FD_CAN frame been received
						//remove the socket
						int close_fd = sock_list[a].fd;
						std::cout
								<< "close_fd, due to read: invalid CAN / CAN FD frame"
								<< close_fd << std::endl;
						sock_list[a].status = 0;
						sock_list[a].protocol_family = 0;
						num_of_sock--;
						Handler_Remove_Fd_Read(close_fd);
						close(close_fd);
					}
				}
				}
			}
			case CAN_BCM: {
				SocketCAN__Types::SocketCAN__receive__BCM__message parameters;
				struct sockaddr_can addr;
				socklen_t len = sizeof(addr);
				struct {
					struct bcm_msg_head msg_head;
					struct canfd_frame frame[BCM_FRAME_BUFFER_SIZE];
				} bcm_msg;

				struct ifreq ifr;
				ifr.ifr_ifindex = addr.can_ifindex;
				/* get interface name of the received CAN frame */
				struct ifreq ifr;
				ifr.ifr_ifindex = addr.can_ifindex;
				/* get interface name of the received CAN frame */

				//ssize_t recvfrom(int sockfd, void *buf, size_t len, int flags,
				// struct sockaddr *src_addr, socklen_t *addrlen);
				ssize_t nbytes = recvfrom(sock, &bcm_msg, sizeof(bcm_msg), 0,
						(struct sockaddr*) &addr, &len);
				if (nbytes <= 0) {
					//there is an empty message, or error in receive
					//remove the socket
					TTCN_warning(
							"Closing socket % with interface index %d due to an empty BCM message or error in reception\n",
							sock, addr.can_ifindex);
					std::cout << "close_fd" << sock << std::endl;
					sock_list[a].status = 0;
					sock_list[a].protocol_family = 0;
					num_of_sock--;
					Handler_Remove_Fd_Read (close_fd);
					close(sock);
				} else {
					if (res = ioctl(sock, SIOCGIFNAME, &ifr) < 0) {
						TTCN_warning(
								"Ioctl failed while receiving a BCM message frame on socket: %d with interface index %d\n",
								sock, ifr.ifr_ifindex);
					}
					if (debugging) { //if debug
						TTCN_Logger::begin_event(TTCN_DEBUG);
						TTCN_Logger::log_event_str(prompt);
						TTCN_Logger::log_event("Received a BCM message from interface %s of bytes %d", ifr.ifr_name, nbytes);
						TTCN_Logger::log_event("\n");
						TTCN_Logger::end_event();
					}
					parameters.id() = bcm_msg.msg_head.can_id;
					parameters.ifr().if__index() = ifr.ifr_ifindex;
					parameters.ifr().if__name() = ifr.ifr_name;
					long count = bcm_msg.msg_head.count;
					long flags = bcm_msg.msg_head.flags;
					long tv_sec1 = bcm_msg.msg_head.ival1.tv_sec;
					long tv_usec1 = bcm_msg.msg_head.ival1.tv_usec;
					long tv_sec2 = bcm_msg.msg_head.ival2.tv_sec;
					long tv_usec2 = bcm_msg.msg_head.ival2.tv_usec;
					int nframes = bcm_msg.msg_head.nframes;
					parameters.frame().opcode() = bcm_msg.msg_head.opcode;
					parameters.frame().count() = bcm_msg.msg_head.count;
					parameters.frame().ival1().tv__sec() =
							bcm_msg.msg_head.ival1.tv_sec;
					parameters.frame().ival1().tv__usec() =
							bcm_msg.msg_head.ival1.tv_usec;
					parameters.frame().ival2().tv__sec() =
							bcm_msg.msg_head.ival2.tv_sec;
					parameters.frame().ival2().tv__usec() =
							bcm_msg.msg_head.ival2.tv_usec;
					parameters.frame().can__id() = bcm_msg.msg_head.can_id;

					if ((flags & CAN_FD_FRAME) == CAN_FD_FRAME) {
						// Handle CAN FD frames
						for (uint i = 0;
								i < std::min(nframes, BCM_FRAME_BUFFER_SIZE);
								i++) {
							parameters.frame().frames().canfd__frame()[i].can__id() =
									bcm_msg.frame[i].can_id;
							//Here the bitstring shall be stored into a
							parameters.frame().frames().canfd__frame()[i].can__flags() =
									bcm_msg.frame[i].flags;

							int len = bcm_msg.frame[i].len;
							for (uint j = 0; j < std::min(len, CANFD_MAX_DLEN);
									j++) {
								parameters.frame().frames().canfd__frame()[j].can__pdu()[j] =
										bcm_msg.msg_head.frames[i].data[j];
							}

						}
					} else {
						// Handle legacy CAN frames
						for (uint i = 0;
								i < std::min(nframes, BCM_FRAME_BUFFER_SIZE);
								i++) {
							parameters.frame().frames().can__frame()[i].can__id() =
									bcm_msg.msg_head.frames[i].can_id;
							int len = bcm_msg.frame[i].len;
							for (uint j = 0; j < std::min(len, CANFD_MAX_DLEN);
									j++) {
								parameters.frame().frames().can__frame()[j].can__pdu()[j] =
										bcm_msg.msg_head.frames[i].data[j];
							}
						}
					}
				}
			}
				break;
			default: {
				TTCN_error(
						"SocketCAN Handle_Fd_Event_Readable (%d): unhandled protocol configured",
						sock);
			}
				break;
			}
		}
	}
	log("leaving SocketCAN__PT_PROVIDER::Handle_Fd_Event_Readable()");
}

void SocketCAN__PT_PROVIDER::user_map(const char *system_port) {
	log("entering SocketCAN__PT_PROVIDER::user_map()");

	if (sock_list != NULL)
		TTCN_error("SocketCAN Test Port (%s): Internal error: "
				"sock_list is not NULL when mapping.", port_name);
	sock_list = (sock_data*) Malloc(DEFAULT_NUM_SOCK * sizeof(*sock_list));
	num_of_sock = 0;
	sock_list_length = DEFAULT_NUM_SOCK;
	for (int a = 0; a < sock_list_length; a++) {
		sock_list[a].status = 0;
		sock_list[a].protocol_family = 0;
	}

	log("leaving SocketCAN__PT_PROVIDER::user_map()");
}

void SocketCAN__PT_PROVIDER::user_unmap(const char * /*system_port*/) {
	log("entering SocketCAN__PT_PROVIDER::user_unmap()");

	closeDownSocket();

	log("leaving SocketCAN__PT_PROVIDER::user_unmap()");
}

void SocketCAN__PT_PROVIDER::user_start() {

}

void SocketCAN__PT_PROVIDER::user_stop() {

}

void SocketCAN__PT_PROVIDER::outgoing_send(
		const SocketCAN__Types::SocketCAN__socket& send_par) {
	log("entering SocketCAN__PT_PROVIDER::outgoing_send(SocketCAN__socket)");
	int cn;
	SocketCAN__Types::SocketCAN__socket__result result;

	if (num_of_sock < sock_list_length) {
		cn = 0;
		while (sock_list[cn].status) {
			cn++;
		}
	} else {
		sock_list = (sock_data*) Realloc(sock_list,
				2 * sock_list_length * sizeof(*sock_list));
		for (int a = sock_list_length; a < sock_list_length * 2; a++) {
			sock_list[a].status = 0;
		}
		cn = sock_list_length;
		sock_list_length *= 2;
	}

//extern int socket (int __domain, int __type, int __protocol) __THROW;
	if ((target_fd = socket(send_par.domain(), send_par.ptype(),
			send_par.protocol())) < 0) {
		TTCN_warning("Cannot open socket \n");
		result.result()().result__code() =
				SocketCAN__Types::SocketCAN__Result__code::ERROR_;
	}

	else {

		sock_list[cn].fd = target_fd;
		sock_list[cn].status = 1;

		num_of_sock++;

		//Handler_Add_Fd_Read(target_fd);

		result.id() = cn;
		result.result()().result__code() =
				SocketCAN__Types::SocketCAN__Result__code::SUCCESS;
		result.result()().err() = OMIT_VALUE;
	}

	incoming_message(result);
	log("leaving SocketCAN__PT_PROVIDER::outgoing_send(SocketCAN__socket)");
}

void SocketCAN__PT_PROVIDER::outgoing_send(
		const SocketCAN__Types::SocketCAN__ioctl& send_par) {
	log("entering SocketCAN__PT_PROVIDER::outgoing_send(SocketCAN__ioctl)");

	struct ifreq ifr;

	int sock;
	int cn = send_par.id();
	int res;
	SocketCAN__Types::SocketCAN__ioctl__result result;

	if ((cn < sock_list_length) and (sock_list[cn].status != 0)) {
		if (send_par.ifu().if__name().lengthof() <= IFNAMSIZ) {
			sock = sock_list[cn].fd;

			SocketCAN__ioctl::union_selection_type selection =
					send_par.ifu().get_selection();

			switch (selection) {
			case send_par.ifu().ALT_if__name:
				strcpy(ifr.ifr_name, send_par.ifu().if__name());
				if ((res = ioctl(sock, SIOCGIFINDEX, &ifr)) < 0) {
					TTCN_warning(
							"Ioctl failed on socket: %d with interface name %s\n",
							sock, send_par.ifu().if__name());
					result.ifr().if__name() = send_par.ifu().if__name();
					result.result()().result__code() =
							SocketCAN__Types::SocketCAN__Result__code::ERROR_;

				} else {
					result.ifr().if__name() = send_par.ifu().if__name();
					result.ifr().if__index() = ifr.ifr_ifindex;
					result.result()().result__code() =
							SocketCAN__Types::SocketCAN__Result__code::SUCCESS;
					result.result()().err() = OMIT_VALUE;
				}

				break;
			case send_par.ifu().ALT_if__index:
				ifr.ifr_ifindex = send_par.ifu().if__index();
				if ((res = ioctl(sock, SIOCGIFNAME, &ifr)) < 0) {
					TTCN_warning(
							"Ioctl failed on socket: %d with interface index %s\n",
							sock, send_par.ifu().if__index());
					result.ifr().if__index() = ifr.ifr_ifindex;
					result.result()().result__code() =
							SocketCAN__Types::SocketCAN__Result__code::ERROR_;
				} else {
					result.ifr().if__name() = ifr.ifr_name;
					result.result()().result__code() =
							SocketCAN__Types::SocketCAN__Result__code::SUCCESS;
					result.result()().err() = OMIT_VALUE;
				}
				break;
			default:
				TTCN_error("Ioctl failed due to unknown union selection");
				break;
			}
		} else {
			TTCN_error("Ioctl failed due to interface name too long.\n");
			result.ifr().if__name() = ifr.ifr_name;
			result.result()().result__code() =
					SocketCAN__Types::SocketCAN__Result__code::ERROR_;
		}
	} else {
		TTCN_error("Ioctl failed due to unknown socket reference: %d \n", cn);
		result.ifr().if__name() = ifr.ifr_name;
		result.result()().result__code() =
				SocketCAN__Types::SocketCAN__Result__code::ERROR_;
	}
	incoming_message(result);
	log("SocketCAN__PT::outgoing_send(SocketCAN__ioctl)");
}
void SocketCAN__PT_PROVIDER::outgoing_send(
		const SocketCAN__Types::SocketCAN__connect& send_par) {
//Client connects to BCM
	log("entering SocketCAN__PT_PROVIDER::outgoing_send(SocketCAN__connect)");

	int sock;
	struct sockaddr_can addr;
	int cn = send_par.id();
	int res;
	SocketCAN__Types::SocketCAN__connect__result result;

	if ((cn < sock_list_length) and (sock_list[cn].status != 1)) {
		if (sock_list[cn].protocol_family != 0) {
			sock = sock_list[cn].fd;

			addr.can_family = AF_CAN;
			addr.can_ifindex = send_par.if__index();

			//extern int connect (int __fd, __CONST_SOCKADDR_ARG __addr, socklen_t __len);
			if (connect(sock, (struct sockaddr *) &addr, sizeof(addr)) != 0) {
				result.result()().result__code() =
						SocketCAN__Types::SocketCAN__Result__code::ERROR_;
				result.result()().err() = "Connect was unsuccessful";
			} else {
				sock_list[cn].protocol_family = CAN_BCM;
				result.result()().result__code() =
						SocketCAN__Types::SocketCAN__Result__code::SUCCESS;
				result.result()().err() = OMIT_VALUE;
				sock_list[cn].remote_Addr.can_family = AF_CAN;
				Handler_Add_Fd_Read(target_fd);
			}
		} else {
			TTCN_error("Socket reference already connected or bound: %d \n",
					cn);
			result.result()().result__code() =
					SocketCAN__Types::SocketCAN__Result__code::ERROR_;
		}
	} else {
		TTCN_error("Unknown socket reference: %d \n", cn);
		result.result()().result__code() =
				SocketCAN__Types::SocketCAN__Result__code::ERROR_;
	}
	incoming_message(result);
	log("leaving SocketCAN__PT_PROVIDER::outgoing_send(SocketCAN__connect)");
}

void SocketCAN__PT_PROVIDER::outgoing_send(
		const SocketCAN__Types::SocketCAN__bind& send_par) {
//Client binds
	log("entering SocketCAN__PT_PROVIDER::outgoing_send(SocketCAN__bind)");

	int sock;
	struct sockaddr_can addr;
	int cn = send_par.id();
	int res;
	SocketCAN__Types::SocketCAN__bind__result result;

	if ((cn < sock_list_length) and (sock_list[cn].status != 1)) {
		if (sock_list[cn].protocol_family != 0) {
			sock = sock_list[cn].fd;

			addr.can_family = AF_CAN;
			addr.can_ifindex = send_par.if__index();

			if (bind(sock, (struct sockaddr *) &addr, sizeof(addr)) != 0) {
				result.result()().result__code() =
						SocketCAN__Types::SocketCAN__Result__code::ERROR_;
				TTCN_error("Bind was unsuccessful: %d \n", sock);
				result.result()().err() = "Bind was unsuccessful";
			} else {
				sock_list[cn].protocol_family = CAN_RAW;
				result.result()().result__code() =
						SocketCAN__Types::SocketCAN__Result__code::SUCCESS;
				result.result()().err() = OMIT_VALUE;
				sock_list[cn].remote_Addr.can_family = AF_CAN;
				Handler_Add_Fd_Read(target_fd);
			}
		} else {
			TTCN_error("Socket reference already connected or bound: %d \n",
					cn);
			result.result()().result__code() =
					SocketCAN__Types::SocketCAN__Result__code::ERROR_;
		}
	} else {
		TTCN_error("Unknown socket reference: %d \n", cn);
		result.result()().result__code() =
				SocketCAN__Types::SocketCAN__Result__code::ERROR_;
	}
	incoming_message(result);
	log("leaving SocketCAN__PT_PROVIDER::outgoing_send(SocketCAN__bind)");
}

void SocketCAN__PT_PROVIDER::outgoing_send(
		const SocketCAN__Types::SocketCAN__send__data& send_par) {
	log(
			"entering SocketCAN__PT_PROVIDER::outgoing_send(SocketCAN__send__data)");

	SocketCAN__Types::SocketCAN__send__data__result result;
	int res;
	int sock;
	int cn = send_par.id();
	int nrOfBytesSent, nrOfBytestoSent;

	SocketCAN__Types::SocketCAN__send__data::union_selection_type can_frame_type_selection =
			send_par.frame().get_selection();

	if ((cn < sock_list_length)) {
		struct sockaddr_can addr;
		struct ifreq ifr;
		if (send_par.ifu().ispresent()) {
			switch (send_par.ifu().get_selection()) {
			case SocketCAN__ifu::ALT_if__name:
				strcpy(ifr.ifr_name, send_par.ifu().if__name());
				sock = sock_list[cn].fd;
				if (res = ioctl(sock, SIOCGIFINDEX, &ifr) < 0) {
					TTCN_warning(
							"Send CAN fame: Ioctl failed while retrieving the interface : %d with interface index %d\n",
							sock, ifr.ifr_name);
				}
			case SocketCAN__ifu::ALT_if__index:
				addr.can_ifindex = ifr.ifr_ifindex;
				addr.can_family = AF_CAN;

				break;
			default:
				TTCN_warning("Send CAN fame: Unknown union selection");
			}
		}

		socklen_t len = sizeof(addr);
		SocketCAN__Types::SocketCAN__send__data::union_selection_type can_frame_type_selection =
				send_par.frame().get_selection();

		switch (can_frame_type_selection) {
		case send_par.frame().ALT_can__frame:
			struct can_frame frame;

			log("Sending CAN frame)");
			logHex(" to can id: ", send_par.frame().can__frame().can__id());
			logHex("containing data: ",
					send_par.frame().can__frame().can__pdu());

			frame.can_id = send_par.frame().can__frame().can__id();
			memcpy(frame.data, send_par.frame().can__frame().can__pdu(),
					send_par.frame().can__frame().can__pdu().lengthof());
			frame.can_dlc = send_par.frame().can__frame().can__pdu().lengthof();

			nrOfBytestoSent = sizeof(frame);
			if (send_par.ifu().ispresent()) {
				nrOfBytesSent = sendto(sock, &frame, nrOfBytestoSent, 0,
						(struct sockaddr*) &addr, sizeof(addr));
			} else {
				nrOfBytesSent = send(sock, &frame, nrOfBytestoSent, 0);
			}
			if (nrOfBytesSent <= 0) {
				TTCN_error("SocketCAN CAN send error");
				result.result()().result__code() =
						SocketCAN__Types::SocketCAN__Result__code::ERROR_;
			}
			break;
		case send_par.frame().ALT_can__fd__frame:
			struct canfd_frame fd_frame;

			log("Sending CAN FD frame)");
			logHex(" to can id: ", send_par.frame().can__fd__frame().can__id());
			logBitstring("with flags: ",
					send_par.frame().can__fd__frame().can__flags());
			logHex("containing data: ",
					send_par.frame().can__fd__frame().can__pdu());

			fd_frame.can_id = send_par.frame().can__fd__frame().can__id();
			memcpy(fd_frame.data, send_par.frame().can__fd__frame().can__pdu(),
					send_par.frame().can__fd__frame().can__pdu().lengthof());
			fd_frame.len =
					send_par.frame().can__fd__frame().can__pdu().lengthof();

			nrOfBytestoSent = sizeof(fd_frame);
			if (send_par.ifu().ispresent()) {
				nrOfBytesSent = sendto(sock, &fd_frame, nrOfBytestoSent, 0,
						(struct sockaddr*) &addr, sizeof(addr));
			} else {
				nrOfBytesSent = send(sock, &fd_frame, nrOfBytestoSent, 0);
			}
			if (nrOfBytesSent <= 0) {
				TTCN_error("SocketCAN CAN FD send error");
				result.result()().result__code() =
						SocketCAN__Types::SocketCAN__Result__code::ERROR_;
			}
			result.result()().result__code() =
					SocketCAN__Types::SocketCAN__Result__code::ERROR_;
			break;
		default:
			TTCN_error("SocketCAN send unknown frame type error");
		}

		log("Nr of bytes sent = %d", nrOfBytesSent);

		if (nrOfBytesSent != nrOfBytestoSent) {
			TTCN_error(
					"send system call failed: %d bytes was sent instead of %d",
					nrOfBytesSent, nrOfBytestoSent);
			result.result()().result__code() =
					SocketCAN__Types::SocketCAN__Result__code::ERROR_;
		}
	} else {
		TTCN_error(
				"send system call failed due to unknown socket reference: %d \n",
				cn);
		result.result()().result__code() =
				SocketCAN__Types::SocketCAN__Result__code::ERROR_;
	}
	incoming_message(result);
	log("leaving SocketCAN__PT_PROVIDER::outgoing_send(SocketCAN__send__data)");
}

void SocketCAN__PT_PROVIDER::outgoing_send(
		const SocketCAN__Types::SocketCAN__write__data& send_par) {
	log(
			"entering SocketCAN__PT_PROVIDER::outgoing_send(SocketCAN__write__data)");

	SocketCAN__Types::SocketCAN__write__data__result result;
	int sock;
	int nrOfBytesSent, nrOfBytestoSent;

	// unsure whether here need can fd frames to be supported too

	struct {
		struct bcm_msg_head msg_head;
		struct can_frame frame[BCM_FRAME_BUFFER_SIZE];
	} bcm_msg;

	SocketCAN__Types::SocketCAN__write__data::union_selection_type can_frame_type_selection =
			send_par.frame().get_selection();

	bcm_msg.msg_head.nframes = send_par.bcm__tx__msg().frames()

	if (send_par.bcm__tx__msg().frames().can__frame().lengthof() > BCM_FRAME_BUFFER_SIZE) {
		TTCN_error("Writing data: number of frames too large: %d \n");
	}

	switch (can_frame_type_selection) {
	case ALT_can__frame:
		int nframes = std::min(
				send_par.bcm__tx__msg().frames().can__frame().lengthof(),
				BCM_FRAME_BUFFER_SIZE);
		bcm_msg.msg_head.opcode = send_par.bcm__tx__msg().opcode();
		bcm_msg.msg_head.flags = send_par.bcm__tx__msg().flags();
		bcm_msg.msg_head.count = send_par.bcm__tx__msg().count();
		bcm_msg.msg_head.ival1.tv_sec = send_par.bcm__tx__msg().ival1().tv__sec();
		bcm_msg.msg_head.ival1.tv_usec = send_par.bcm__tx__msg().ival1().tv__usec();
		bcm_msg.msg_head.ival2.tv_sec = send_par.bcm__tx__msg().ival2().tv__sec();
		bcm_msg.msg_head.ival2.tv_usec = send_par.bcm__tx__msg().ival2().tv__usec();
		bcm_msg.msg_head.nframes = nframes;

		log("Sending BCM Message)");
		logHex(" opcode: ", send_par.bcm__tx__msg().opcode());
		logBitstring(" flags: ", send_par.bcm__tx__msg().flags());
		logHex(" count: ", send_par.bcm__tx__msg().count());
		logInteger(" ival1: ", send_par.bcm__tx__msg().ival1().tv__sec());
		logInteger(" ival1: ", send_par.bcm__tx__msg().ival1().tv__usec());
		logInteger(" ival2: ", send_par.bcm__tx__msg().ival2().tv__sec());
		logInteger(" ival2: ", send_par.bcm__tx__msg().ival2().tv__usec());
		logHex(" can_id: ", send_par.bcm__tx__msg().can__id());
		logHex(" nframes: ", nframes);

		for (uint i = 0; i < nframes; i++) {
			bcm_msg.frame[i].can_id =
					send_par.bcm__tx__msg().frames().can__frame()[i].can__id();
			bcm_msg.frame[i].can_dlc = nframes;
			if (send_par.bcm__tx__msg().frames().can__frame()[i].can__pdu().lengthof() > CAN_MAX_DLEN) {
				TTCN_error("Writing data: frame has too much pdu data\n");
			}

			log(" containing with CAN frame:)");
			logHex("   can id: ", bcm_msg.frame[i].can_id);
			logHex("   can dlc: ", bcm_msg.frame[i].can_dlc);

			int len =
					std::min(
							send_par.bcm__tx__msg().frames().can__frame()[i].can__pdu().lengthof(),
							CAN_MAX_DLEN);

			bcm_msg.frame[i].can_dlc = len;
			for (uint j = 0; j < len; j++) {
				bcm_msg.frame[i].data[j] =
						send_par.bcm__tx__msg().frames().can__frame()[i].can__pdu()[j];
				logHex("   data: ", bcm_msg.frame[i].data[j]);
			}
			nrOfBytestoSent = sizeof(bcm_msg);
			nrOfBytesSent = write(sock, &bcm_msg, nrOfBytestoSent);

			if (nrOfBytesSent <= 0) {
				TTCN_error("SocketCAN CAN send error");
				result.result()().result__code() =
						SocketCAN__Types::SocketCAN__Result__code::ERROR_;
			}
		}

	case ALT_canfd__frame:
		// not supported, unclear if needed
		TTCN_error("Writing data: sending of CAN FD frames is not supported");
		break;

	default:
		TTCN_error("SocketCAN send unknown frame type error");
	}

	log("Nr of bytes sent = %d", nrOfBytesSent);

	if (nrOfBytesSent != nrOfBytestoSent) {
		TTCN_error("write system call failed: %d bytes was sent instead of %d",
				nrOfBytesSent, nrOfBytestoSent);
		result.result()().result__code() =
				SocketCAN__Types::SocketCAN__Result__code::ERROR_;
	} else {
		TTCN_error(
				"Write system call failed due to unknown socket reference: %d \n",
				cn);
		//result.ifr.ifr__name() = send_par.if__name();
		result.result()().result__code() =
				SocketCAN__Types::SocketCAN__Result__code::ERROR_;
	}
	incoming_message(result);
	log(
			"leaving SocketCAN__PT_PROVIDER::outgoing_send(SocketCAN__write__data)");
}

void SocketCAN__PT_PROVIDER::outgoing_send(
		const SocketCAN__Types::SocketCAN__setsockopt& send_par) {
	log(
			"entering SocketCAN__PT_PROVIDER::outgoing_send(SocketCAN__setsockopt)");

	struct ifreq ifr;

	int sock;
	int cn = send_par.id();
	int res;
	SocketCAN__Types::SocketCAN__setsockopt__result result;

	if ((cn < sock_list_length) and (sock_list[cn].status != 0)) {
		sock = sock_list[cn].fd;

		SocketCAN__setsockopt::union_selection_type command_selection =
				send_par.command().get_selection();

		switch (command_selection) {
		case send_par.command().ALT_set__filters: {

			std::size_t rfilter_size =
					(sizeof(send_par.command().set__filters().rfilter())
							/ sizeof(send_par.command().set__filters().rfilter()[0]));

			struct can_filter rfilter[rfilter_size];

			SocketCAN__setsockopt::union_selection_type filter_selection =
					send_par.command().set__filters().get_selection();

			switch (filter_selection) {
			case send_par.command().set__filters().ALT_rfilter: {
				for (int i = 0; i < rfilter_size; i++) {
					rfilter[i].can_id =
							send_par.command().set__filters().rfilter().can__id();
					rfilter[i].can_mask =
							send_par.command().set__filters().rfilter().can__mask();
				};
				if (res = setsockopt(sock, SOL_CAN_RAW, CAN_RAW_FILTER,
						&rfilter, sizeof(rfilter)) < 0) {
					result.result()().result__code() =
							SocketCAN__Types::SocketCAN__Result__code::ERROR_;
					result.result()().err() = "Setsockopt was unsuccessful";
				} else {
					result.result()().result__code() =
							SocketCAN__Types::SocketCAN__Result__code::SUCCESS;
					result.result()().err() = OMIT_VALUE;
				}
			}
				break;
				;
			case send_par.command().set__filters().ALT_filter: {
				if (res = setsockopt(sock, SOL_CAN_RAW, CAN_RAW_FILTER, NULL, 0)
						< 0) {
					result.result()().result__code() =
							SocketCAN__Types::SocketCAN__Result__code::ERROR_;
					result.result()().err() = "Setsockopt was unsuccessful";
				} else {
					result.result()().result__code() =
							SocketCAN__Types::SocketCAN__Result__code::SUCCESS;
					result.result()().err() = OMIT_VALUE;
				}
			}
				break;

			}
		}
			break;

		case send_par.command().ALT_can__err__mask: {

			can_err_mask_t err_mask = send_par.command().can__err__mask();

			if (res = setsockopt(sock, SOL_CAN_RAW, CAN_RAW_ERR_FILTER,
					&err_mask, sizeof(err_mask)) < 0) {
				result.result()().result__code() =
						SocketCAN__Types::SocketCAN__Result__code::ERROR_;
				result.result()().err() = "Setsockopt was unsuccessful";
			} else {
				result.result()().result__code() =
						SocketCAN__Types::SocketCAN__Result__code::SUCCESS;
				result.result()().err() = OMIT_VALUE;
			}
		}
			break;

		case send_par.command().ALT_loopback: {

			int loopback = send_par.command().loopback();

			if (res = setsockopt(sock, SOL_CAN_RAW, CAN_RAW_LOOPBACK, &loopback,
					sizeof(loopback)) < 0) {
				result.result()().result__code() =
						SocketCAN__Types::SocketCAN__Result__code::ERROR_;
				result.result()().err() = "Setsockopt was unsuccessful";
			} else {
				result.result()().result__code() =
						SocketCAN__Types::SocketCAN__Result__code::SUCCESS;
				result.result()().err() = OMIT_VALUE;
			}
		}
			break;

		case send_par.command().ALT_recv__own__msg: {

			int recv_own_msgs = send_par.command().recv__own__msg();

			if (res = setsockopt(sock, SOL_CAN_RAW, CAN_RAW_RECV_OWN_MSGS,
					&recv_own_msgs, sizeof(recv_own_msgs)) < 0) {
				result.result()().result__code() =
						SocketCAN__Types::SocketCAN__Result__code::ERROR_;
				result.result()().err() = "Setsockopt was unsuccessful";
			} else {
				result.result()().result__code() =
						SocketCAN__Types::SocketCAN__Result__code::SUCCESS;
				result.result()().err() = OMIT_VALUE;
			}
		}
			break;

		case send_par.command().ALT_fd__frames: {

			int fd_frames = send_par.command().fd__frames();

			if (res = setsockopt(sock, SOL_CAN_RAW, CAN_RAW_FD_FRAMES,
					&fd_frames, sizeof(fd_frames)) < 0) {
				result.result()().result__code() =
						SocketCAN__Types::SocketCAN__Result__code::ERROR_;
				result.result()().err() = "Setsockopt was unsuccessful";
			} else {
				result.result()().result__code() =
						SocketCAN__Types::SocketCAN__Result__code::SUCCESS;
				result.result()().err() = OMIT_VALUE;
			}
		}
			break;
		case send_par.command().ALT_join__filters: {

			int join_filters = send_par.command().join__filters();

			if (res = setsockopt(sock, SOL_CAN_RAW, CAN_RAW_JOIN_FILTERS,
					&join_filters, sizeof(join_filters)) < 0) {
				result.result()().result__code() =
						SocketCAN__Types::SocketCAN__Result__code::ERROR_;
				result.result()().err() = "Setsockopt was unsuccessful";
			} else {
				result.result()().result__code() =
						SocketCAN__Types::SocketCAN__Result__code::SUCCESS;
				result.result()().err() = OMIT_VALUE;
			}
		}
			break;
		}
	} else {
		TTCN_error("Unknown socket reference: %d \n", cn);
		result.result()().result__code() =
				SocketCAN__Types::SocketCAN__Result__code::ERROR_;
	}
}

void SocketCAN__PT_PROVIDER::outgoing_send(
		const SocketCAN__Types::SocketCAN__close& send_par) {
	log("entering SocketCAN__PT_PROVIDER::outgoing_send(SocketCAN__close)");
	int close_fd = sock_list[send_par.id()].fd;

	sock_list[send_par.id()].status = 0;
	sock_list[send_par.id()].fd = 0;
	sock_list[send_par.id()].protocol_family = 0;
	num_of_sock--;
	Handler_Remove_Fd_Read(close_fd);

	close(close_fd);

	log("leaving SocketCAN__PT_PROVIDER::outgoing_send(SocketCAN__close)");
}

void SocketCAN__PT_PROVIDER::log(const char *fmt, ...) {
	if (debugging) {
		TTCN_Logger::begin_event(TTCN_DEBUG);
		TTCN_Logger::log_event("SocketCAN test port (%s): ", get_name());
		va_list args;
		va_start(args, fmt);
		TTCN_Logger::log_event_va_list(fmt, args);
		va_end(args);
		TTCN_Logger::end_event();
	}
}

void SocketCAN__PT_PROVIDER::logHex(const char *prompt,
		const OCTETSTRING& msg) {
	if (debugging) { //if debug
		TTCN_Logger::begin_event(TTCN_DEBUG);
		TTCN_Logger::log_event_str(prompt);
		TTCN_Logger::log_event("Size: %d,\nMsg: ",msg.lengthof());

		for(int i=0; i<msg.lengthof(); i++) {
			TTCN_Logger::log_event(" %02x", ((const unsigned char*)msg)[i]);
		}
		TTCN_Logger::log_event("\n");
		TTCN_Logger::end_event();
	}
}

void SocketCAN__PT_PROVIDER::logInteger(const char *prompt,
		const int number) {
	if (debugging) { //if debug
		TTCN_Logger::begin_event(TTCN_DEBUG);
		TTCN_Logger::log_event_str(prompt);
		TTCN_Logger::log_event("Value: %d,\n: ", number);
		TTCN_Logger::log_event("\n");
		TTCN_Logger::end_event();
	}
}

void SocketCAN__PT_PROVIDER::logBitstring(const char *prompt,
		const BITSTRING& msg) {
	if (debugging) { //if debug
		TTCN_Logger::begin_event(TTCN_DEBUG);
		TTCN_Logger::log_event_str(prompt);
		TTCN_Logger::log_event("Size: %d,\nMsg: ",msg.lengthof());

		for(int i=0; i<msg.lengthof(); i++) {
			TTCN_Logger::log_event(" %02b", ((const unsigned char*)msg)[i]);
		}
		TTCN_Logger::log_event("\n");
		TTCN_Logger::end_event();
	}
}

void SocketCAN__PT_PROVIDER::setUpSocket() {
	log("entering SocketCAN__PT_PROVIDER::setUpSocket()");
	log("leaving SocketCAN__PT_PROVIDER::setUpSocket()");
}

void SocketCAN__PT_PROVIDER::closeDownSocket() {
	log("entering SocketCAN__PT_PROVIDER::closeDownSocket()");

	Free (sock_list_server);
	sock_list_server = NULL;

	for (int a = 0; a < sock_list_length; a++) {
		if (sock_list[a].status) {
			sock_list[a].status = 0;
			sock_list[a].protocol_family = 0;
			close(sock_list[a].fd);
			Handler_Remove_Fd_Read(sock_list[a].fd);
		}
	}

	Free(sock_list);
	sock_list = NULL;

	log("leaving SocketCAN__PT_PROVIDER::closeDownSocket()");
}

} /* end of namespace */

